package GO::Object::Typedef;

=head1 NAME

GO::Object::Typedef       - represents a typedef in an OBO file

=cut

use strict;
use Data::Dumper;
use Exporter;
use GO::TestSet qw(dfv_test);
use lib '/Users/gwg/go/scratch/tools';
use vars qw(@ISA);
use base qw(GO::Object::Base);



sub _specification {
	my $self = shift;
	
	my $spec_h = $self->SUPER::get_spec;
	
	return (
		"id", {
			test => qr/^$self->__id_format()$/,
			required => 1,
		},
		"is_anonymous", {
			test => dfv_test('is_true_p', { this => 1 }),
		},
		"name", {
			test => dfv_test('is_a_string_p', { this => 1 }),
			required => 1,
		},
		"namespace", {
			test => dfv_test('is_a_string_p', { this => 1 }),
		},
		"alt_id", {
			test => $self->__id_format,
			allow_multiple => 1,
		},
		"def", {
			test => dfv_test('is_a_string_p', { this => 1 }),
		},
		"comment", $spec_h->{comment},
		"subset", {
			test => dfv_test('is_a_string_p', { this => 1 }),
			allow_multiple => 1,
		},
		"synonym", {
			test => dfv_test('is_a_string_p', { this => 1 }),
			allow_multiple => 1,
		},
		"xref", {
			test => qr/^.+:.+$/,
			allow_multiple => 1,
		},
		"is_a", {
			test => qr/^$self->__id_format() \! .+/,
			allow_multiple => 1,
		},
		## specific to typedefs
		"domain", {
		},
		"range", {
		},
		"is_anti_symmetric", {
		},
		"is_cyclic", {
		},
		"is_reflexive", {
		},
		"is_symmetric", {
		},
		"is_transitive", {
		},
		"is_a", {
				test => qr/^$self->__id_format() \! .+/,
				allow_multiple => 1,
		},
		"inverse_of", {
		},
		"transitive_over", {
		},
		"relationship", {
				test => qr/ $self->__id_format() \! .+/,
				allow_multiple => 1,
		},
		'is_obsolete', $spec_h->{is_obsolete},
		'replaced_by', $spec_h->{replaced_by},
		'consider', $spec_h->{consider},
	);
}

sub __id_format {
	return qr/GO:\d{7}/;
}




1;