#!/usr/bin/perl -w

use strict;

#use CGI::Test;
use Test::More qw(no_plan);
use Test::WWW::Mechanize::CGIApp;
use lib "/Users/gwg/go/scratch/tools/";
use Carp ();
$SIG{__WARN__} = \&Carp::cluck;
$SIG{__DIE__} = \&Carp::confess;

$ENV{TMPL_VERBOSE} = 1;

use Data::Dumper;
$Data::Dumper::Sortkeys = 1;

use_ok( 'GO::MsgLite' );

my $installation_dir = '/Users/gwg/go/scratch/tools/';  # alter as necessary
my $mech = Test::WWW::Mechanize::CGIApp->new;

run_tests($mech);

exit(0);

sub run_tests {

	# prerequisites
	my @list = qw( GO::Tool::GenericGoTool
	CGI::Carp
	Data::FormValidator::Constraints
	GO::Object::Generic
	GO::TestSet
	GO::Utilities
	GO::GeneralPurposeParser
	URI::Escape);
	
	foreach (@list)
	{	use_ok($_);
	}

	# specifics
	my @list_2 = qw(
		GO::Tool::ReferencesTool

		GO::Object::Reference
		GO::Object::Xref
		GO::Object::XrefAbbr
	);

	## OK, let's start with the References and Xrefs tools.
	## Hopefully they will be easy to test!

	foreach (@list_2)
	{	use_ok($_);
#	{	eval { use $_; };
#		die "Could not 'use' specific module $_! $@\n" if $@;
	}

	my ($content, $parser, $parser_args, $html_display); # initialise these now...

	## let's see if we can parse these files without an error occurring...
	my $ref_db = "/Users/gwg/go/scratch/tools/test_data/GO.references";
#	$self->set_verbosity('superverbose');
	# get the parser ready for action
	$parser_args = {
		file => $ref_db,
		file_separator => "\n\n",
		inner_parser => {
			parser => 'multi_field_tag_value',
			return_record_as_object => 'GO::Object::Reference',
			check_input => 1,
			filter_field => qr/^[^!]/,
			filter_record => qr/^go_ref_id/ism,
		},
	};

	$parser = GO::GeneralPurposeParser->new;
	# read in the file
	my $arr = $parser->parse_from_file($parser_args);

	# tests 21 and 22
	# we should have one error and a ton of refs
	ok(scalar @$arr == 20, "found 20 GO refs");
	ok($parser->has_n_msgs == 1, "Parser has one error message");
#	print STDERR "results: ".Dumper($arr)."\n";

	
	# try first with a known incompatible value
	my $go_ref = 'GO_REF:000008';
	$parser_args->{inner_parser}{filter_record} = [ qr/^(go_ref|alt)_id: $go_ref\b/ism ];
	
	undef $arr;
	$arr = $parser->parse_from_file($parser_args);
	print STDERR "arr, looking for $go_ref: ".Dumper($arr);
	ok(! defined $arr, "Failing to find GO_REF:000008" );

	# known crap synonym
	$go_ref = 'GO_REF:000005';
	$parser_args->{inner_parser}{filter_record} = qr/^(go_ref|alt)_id: $go_ref\b/ism;
	undef $arr;
	$arr = $parser->parse_from_file($parser_args);
	print STDERR "arr, looking for $go_ref: ".Dumper($arr);
	ok(! defined $arr, "Cannot find $go_ref" );

	$parser_args->{inner_parser}{filter_record} = $go_ref;
	undef $arr;
	$arr = $parser->parse_from_file($parser_args);
	ok(! defined $arr, "$go_ref search fails" );

	# known compatible value
	$go_ref = 'GO_REF:0000008';
	my $different_args = {
		as_string => $go_ref,
		regexp => qr/^(go_ref|alt)_id: $go_ref\b/ism,
		regexp_in_arr => [ qr/^(go_ref|alt)_id: $go_ref\b/ism ],
		two_regexps_in_arr => [ qr/^go_ref_id: $go_ref/ism, qr/^alt_id: $go_ref/ism,],
		subroutine => sub { my $line = shift; return 1 if $line =~ /^(go_ref|alt)_id: $go_ref/ism; },
		boolean => { OR => [ { FN => sub { my $line = shift; return 1 if $line =~ /^go_ref_id: $go_ref/ism } }, { FN => sub { my $line = shift; return 1 if $line =~ /^alt_id: $go_ref/ism } } ] },
	};

	foreach my $input_type (keys %$different_args)
	{	
		$parser_args->{inner_parser}{filter_record} = $different_args->{$input_type};
		undef $arr;
		$arr = $parser->parse_from_file($parser_args);
		(my $input_txt = $input_type) =~ s/_/ /g;
		ok(scalar @$arr == 1 && $arr->[0]->go_ref_id eq $go_ref, "Searching for $go_ref, $input_txt");
		delete $parser_args->{inner_parser}{filter_record};
		
		$parser_args->{inner_parser}{if_then_record} = { if => $different_args->{$input_type} };
		undef $arr;
		$arr = $parser->parse_from_file($parser_args);
		ok(scalar @$arr == 1 && $arr->[0]->go_ref_id eq $go_ref, "Searching for $go_ref, $input_txt, if_then");
		delete $parser_args->{inner_parser}{if_then_record};
	}

	# known synonym
	$go_ref = 'GO_REF:0000005';
	$different_args = {
		as_string => $go_ref,
		regexp => qr/^(go_ref|alt)_id: $go_ref\b/ism,
		regexp_in_arr => [ qr/^(go_ref|alt)_id: $go_ref\b/ism ],
		two_regexps_in_arr => [ qr/^go_ref_id: $go_ref/ism, qr/^alt_id: $go_ref/ism,],
		subroutine => sub { my $line = shift; return 1 if $line =~ /^(go_ref|alt)_id: $go_ref/ism; },
		boolean => { OR => [ { FN => sub { my $line = shift; return 1 if $line =~ /^go_ref_id: $go_ref/ism } }, { FN => sub { my $line = shift; return 1 if $line =~ /^alt_id: $go_ref/ism } } ] },
	};

	foreach my $input_type (keys %$different_args)
	{	
		$parser_args->{inner_parser}{filter_record} = $different_args->{$input_type};
		undef $arr;
		$arr = $parser->parse_from_file($parser_args);
		(my $input_txt = $input_type) =~ s/_/ /g;
		ok(scalar @$arr == 1 && $arr->[0]->go_ref_id eq 'GO_REF:0000003', "Searching for $go_ref, alt ID for GO_REF:0000003, $input_txt");
		delete $parser_args->{inner_parser}{filter_record};
		
		$parser_args->{inner_parser}{if_then_record} = { if => $different_args->{$input_type} };
		undef $arr;
		$arr = $parser->parse_from_file($parser_args);
		ok(scalar @$arr == 1 && $arr->[0]->go_ref_id eq 'GO_REF:0000003', "Searching for $go_ref, alt ID for GO_REF:0000003, $input_txt, if_then");
		delete $parser_args->{inner_parser}{if_then_record};
	}

	$mech->app("GO::Tool::ReferencesTool");


	$mech->get_ok("", "Creating and getting GO::Tool::ReferencesTool");
	# check that the page has 20 GO refs
	my $content = $mech->content;
	my $count = ($content =~ s/(<h2 class="id">GO_REF:\d+<\/h2>)//g);
	print STDERR "list: ".Dumper($count);
	# t51
	ok($count == 20, "Found 20 GO REFs");
	
	# t52, t53: get a single GO REF
	$mech->get_ok("?id=GO_REF:0000008", "Get a single GO REF");
	$content = $mech->content;
	$count = ($content =~ s/(<h2 class="id">GO_REF:0000008<\/h2>)//g);
	ok( $count == 1, "Found GO_REF:0000008");

	# t54, t55: get a synonym
	$mech->get("?id=GO_REF:0000005");
	$mech->content_like(qr/alt id.*?GO_REF:0000005/i, "Got a GO ref; checking for the synonym");
	$content = $mech->content;
	$count = ($content =~ s/(<h2 class="id">GO_REF:0000003<\/h2>)//g);
	ok( $count == 1, "Found GO_REF:0000005 as an alias for GO_REF:0000003" );

	$html_display = {
		fatal_msg => '<div class="message fatal">',
		warning_msg => '<div class="message warning">',
		info_msg => '<div class="message info">',
		add_tool_title => 'Submit An Exciting Gene Ontology Tool',
		preview_tool_title => 'Preview Your Gene Ontology Tool Submission',
		submit_tool_title => 'GO Tool Submission Successful',
	};

#	$mech->app(
#		sub {
#			my $app = GO::Tool::ReferencesTool->new();
#			$app->set_verbosity('superverbose');
#			$app->run();
#		});

	# get an invalid ID
	$mech->get_ok("?id=GO_REF:000001", "t56: Getting an invalid ID");
	$mech->content_unlike(qr/<h2 class="id">GO_REF:\d+/i, "t57: Checking that no GO REFs were found");
	$mech->content_contains($html_display->{fatal_msg}, "t58: Checking for a fatal message");
	$mech->content_contains("invalid ID", "t59: Checking that the ID is invalid");

	## look for one valid, one invalid, etc..
	## do these tests for Xrefs Tool too.

}

sub make_url_from_hash {
	my $var_h = shift;
	my $url = join("&", 
		map { 
			my $k = $_;
			if (!ref $var_h->{$k})
			{	$k . "=" . $var_h->{$k}
			}
			else
			{	join "&", map { $k ."=". $_ } @{$var_h->{$k}}
			}
		} keys %$var_h);
	return $url;
}

=cut

args: spec: the abbreviated spec
      mech: the mech
      input: what was used as input, as a hash
    [ use_form: the number of the form to use ]

=cut
sub compare_inputs_to_submitted_values {
	my $args = shift;
	if (! $args->{spec} || ! $args->{mech} || ! $args->{input} )
	{	print STDERR "missing an important parameter!\n" && return undef;
	}

	my $mech_thing = $args->{mech};
	my @form_arr = $mech_thing->forms();

	my $n = $args->{use_form} || 0;
	print STDERR "No forms found!\n" && return undef unless $form_arr[$n];
	
	my @input_arr = $form_arr[$n]->inputs();
	
	print STDERR "No inputs found!\n" && return undef unless scalar @input_arr > 0;
	
	my $input_h;
	foreach (@input_arr)
	{	next if $_->type eq 'submit';
		next if $_->name eq 'rm';
		if ($_->type eq 'hidden')
		{	if ($input_h->{$_->name})
			{	if (! ref $input_h->{$_->name} )
				{	$input_h->{$_->name} = [ $input_h->{$_->name}, $_->value ];
				}
				else
				{	push @{$input_h->{$_->name}}, $_->value;
				}
			}
			else
			{	$input_h->{$_->name} = $_->value;
			}
		}
	}

	foreach (keys %{$args->{spec}{default}})
	{	next if $args->{input}{$_};
		## delete the ones that weren't in our original query
		delete $input_h->{$_};
	}
	print STDERR "input_h: ".Dumper($input_h);
	return $input_h;
}
